"""HTML form handling for web clients.

ClientForm is a Python module for handling HTML forms on the client
side, useful for parsing HTML forms, filling them in and returning the
completed forms to the server.  It has developed from a port of Gisle
Aas' Perl module HTML::Form, from the libwww-perl library, but the
interface is not the same.

RFC 1866: HTML 2.0
RFC 1867: Form-based File Upload in HTML
RFC 2388: Returning Values from Forms: multipart/form-data
HTML 3.2 Specification, W3C Recommendation 14 January 1997 (for ISINDEX)
HTML 4.01 Specification, W3C Recommendation 24 December 1999


Copyright 1998-2000 Gisle Aas.
Copyright 2002 John J. Lee <jjl@pobox.com> (The Python port)

This code is free software; you can redistribute it and/or modify it
under the terms of the MIT License (see the file COPYING included with
the distribution).

"""

try: True
except NameError:
    True = 1
    False = 0

import sys, htmllib, formatter, urllib, urllib2, types, string
from types import StringType
try:
    import UnicodeType
except ImportError:
    UNICODE = False
else:
    UNICODE = True
from urlparse import urlparse, urljoin
from copy import copy

VERSION = "0.0.11"

CHUNK = 1024  # size of chunks fed to parser, in bytes

# This version of urlencode is from my Python 1.5.2 back-port of the
# Python 2.1 CVS maintenance branch of urllib.  It will accept a sequence
# of pairs instead of a mapping -- the 2.0 version only accepts a mapping.
def urlencode(query,doseq=False,):
    """Encode a sequence of two-element tuples or dictionary into a URL query string.

    If any values in the query arg are sequences and doseq is true, each
    sequence element is converted to a separate parameter.

    If the query arg is a sequence of two-element tuples, the order of the
    parameters in the output will match the order of parameters in the
    input.
    """

    if hasattr(query,"items"):
        # mapping objects
        query = query.items()
    else:
        # it's a bother at times that strings and string-like objects are
        # sequences...
        try:
            # non-sequence items should not work with len()
            x = len(query)
            # non-empty strings will fail this
            if len(query) and type(query[0]) != types.TupleType:
                raise TypeError
            # zero-length sequences of all types will get here and succeed,
            # but that's a minor nit - since the original implementation
            # allowed empty dicts that type of behavior probably should be
            # preserved for consistency
        except TypeError:
            ty,va,tb = sys.exc_info()
            raise TypeError, "not a valid non-string sequence or mapping object", tb

    l = []
    if not doseq:
        # preserve old behavior
        for k, v in query:
            k = urllib.quote_plus(str(k))
            v = urllib.quote_plus(str(v))
            l.append(k + '=' + v)
    else:
        for k, v in query:
            k = urllib.quote_plus(str(k))
            if type(v) == types.StringType:
                v = urllib.quote_plus(v)
                l.append(k + '=' + v)
            elif UNICODE and type(v) == types.UnicodeType:
                # is there a reasonable way to convert to ASCII?
                # encode generates a string, but "replace" or "ignore"
                # lose information and "strict" can raise UnicodeError
                v = urllib.quote_plus(v.encode("ASCII","replace"))
                l.append(k + '=' + v)
            else:
                try:
                    # is this a sufficient test for sequence-ness?
                    x = len(v)
                except TypeError:
                    # not a sequence
                    v = urllib.quote_plus(str(v))
                    l.append(k + '=' + v)
                else:
                    # loop over the sequence
                    for elt in v:
                        l.append(k + '=' + urllib.quote_plus(str(elt)))
    return string.join(l, '&')


# XXX don't really want to drag this along
class MapBase:
    """Mapping designed to be easily derived from.

    Subclass it and override __init__, __setitem__, __getitem__, __delitem__
    and keys.  Nothing else should need to be overridden, unlike UserDict.
    This significantly simplifies dictionary-like classes.

    Also different from UserDict in that it has a mutable flag, and can be
    updated (and initialised) with a sequence of pairs (key, value).

    """
    def __init__(self, init=None):
        self._data = {}
        self.mutable = 1  # XXX should be inverse flag, and named 'readonly'
        if init is not None: self.update(init)

    def __getitem__(self, key):
        return self._data[key]

    def __setitem__(self, key, item):
        if self.mutable:
            self._data[key] = item
        else:
            raise TypeError, "object doesn't support item assignment"

    def __delitem__(self, key):
        if self.mutable:
            del self._data[key]
        else:
            raise TypeError, "object doesn't support item deletion"

    def keys(self):
        return self._data.keys()

    # now the internal workings, there should be no need to override these:

    def clear(self):
        for k in self.keys():
            del self[k]

    def __repr__(self):
        rep = []
        for k, v in self.items():
            rep.append("%s: %s" % (repr(k), repr(v)))
        return self.__class__.__name__+"{"+(", ".join(rep))+"}"

    def copy(self):
        return copy(self)

    def __cmp__(self, dict):
        # note: return value is *not* boolean
        for k, v in self.items():
            if not (dict.has_key(k) and dict[k] == v):
                return 1  # different
        return 0  # the same

    def __len__(self):
        return len(self.keys())

    def values(self):
        r = []
        for k in self.keys():
            r.append(self[k])
        return r

    def items(self):
        keys = self.keys()
        vals = self.values()
        r = []
        for i in len(self):
            r.append((keys[i], vals[i]))
        return r

    def has_key(self, key):
        return key in self.keys()

    def update(self, map):
        if issequence(map) and not isinstance(map, StringType):
            items = map
        else:
            items = map.items()
        for tup in items:
            if not isinstance(tup, TupleType):
                raise TypeError, (
                    "MapBase.update requires a map or a sequence of pairs")
            k, v = tup
            self[k] = v

    def get(self, key, failobj=None):
        if key in self.keys():
            return self[key]
        else:
            return failobj

    def setdefault(self, key, failobj=None):
        if not self.has_key(key):
            self[key] = failobj
        return self[key]


class AList(MapBase):
    """Read-only ordered mapping."""
    def __init__(self, seq=[]):
        self.mutable = False
        self._inverted = False
        self._data = list(seq[:])
        self._keys = []
        self._values = []
        for key, value in seq:
            self._keys.append(key)
            self._values.append(value)

    def set_inverted(self, inverted):
        if (inverted and not self._inverted) or (
            not inverted and self._inverted):
            self._keys, self._values = self._values, self._keys
        if inverted: self._inverted = True
        else: self._inverted = False

    def __getitem__(self, key):
        try:
            i = self._keys.index(key)
        except ValueError:
            raise KeyError, key
        return self._values[i]

    def __delitem__(self, key):
        try:
            i = self._keys.index[key]
        except ValueError:
            raise KeyError, key
        del self._values[i]

    def keys(self): return list(self._keys[:])
    def values(self): return list(self._values[:])
    def items(self):
        data = self._data[:]
        if not self._inverted:
            return data
        else:
            newdata = []
            for k, v in data:
                newdata.append((v, k))
            return newdata


def issequence(object):
    """Test whether object is a sequence."""
    try:
        object[0]
    except (TypeError, KeyError):
        return False
    except IndexError:
        pass
    return True

def isstringlike(x):
    try: x+""
    except: return False
    else: return True


class ParseError(Exception): pass

def ParseResponse(response, select_default=False, ignore_errors=False):
    """Parse HTTP response and return a list of HTMLForm instances.

    The return value of urllib2.urlopen can be conveniently passed to this
    function as the response parameter.

    ClientForm.ParseError is raised on parse errors.

    response: file-like object (supporting read() method) with a method
     geturl(), returning the base URI of the HTTP response
    select_default: for multiple-selection SELECT controls and RADIO controls,
     pick the first item as the default if none are selected in the HTML
    ignore_errors: don't raise ParseError, and carry on regardless if the
     parser gets confused

    Pass a true value for select_default if you want the behaviour specified by
    RFC 1866 (the HTML 2.0 standard), which is to select the first item in a
    RADIO or multiple-selection SELECT control if none were selected in the
    HTML.  Most browsers (including Microsoft Internet Explorer (IE) and
    Netscape Navigator) instead leave all items unselected in these cases.  The
    W3C HTML 4.0 standard leaves this behaviour undefined in the case of
    multiple-selection SELECT controls, but insists that at least one RADIO
    button should be checked at all times, in contradiction to browser
    behaviour.

    Precisely what ignore_errors does isn't well-defined yet, so don't rely too
    much on the current behaviour -- if you want robustness, you're better off
    fixing the HTML before passing it to this function.

    """
    return ParseFile(response, response.geturl(),
                     select_default, ignore_errors)

def ParseFile(file, base_uri, select_default=False, ignore_errors=False):
    """Parse HTML and return a list of HTMLForm instances.

    ClientForm.ParseError is raised on parse errors.

    file: file-like object (supporting read() method) containing HTML with zero
     or more forms to be parsed
    base_uri: the base URI of the document

    For the other arguments and further details, see ParseResponse.__doc__.

    """
    fp = _FormParser(ignore_errors)
    while 1:
        data = file.read(CHUNK)
        fp.feed(data)
        if len(data) != CHUNK: break
    forms = []
    for (name, action, method, enctype), attrs, controls in fp.forms:
        if action is None:
            action = base_uri
        else:
            action = urljoin(base_uri, action)
        form = HTMLForm(action, method, enctype, name, attrs)
        for type, attr in controls:
            form.new_control(type, attr, select_default=select_default)
        forms.append(form)
    for form in forms:
        form.fixup()
    return forms

class _FormParser(htmllib.HTMLParser):
    # pinched (and modified) from Moshe Zadka
    def __init__(self, ignore_errors=False):
        htmllib.HTMLParser.__init__(self, formatter.NullFormatter())
        self.forms = []
        self._current_form = None
        self._select = None
        self._option = None
        self._textarea = None
        self._ignore_errors = ignore_errors

    def error(self, error):
        if not self._ignore_errors: raise error

    def start_form(self, attrs):
        if self._current_form is not None:
            self.error(ParseError("nested FORMs"))
        name = None
        action = None
        enctype = "application/x-www-form-urlencoded"
        method = "GET"
        d = {}
        for key, value in attrs:
            if key == "name":
                name = value
            elif key == "action":
                action = value
            elif key == "method":
                method = string.upper(value)
            elif key == "enctype":
                enctype = value
            else:
                d["key"] = value
        controls = []
        self._current_form = (name, action, method, enctype), d, controls

    def end_form(self):
        if self._current_form is None:
            self.error(ParseError("end of FORM before start"))
        self.forms.append(self._current_form)
        self._current_form = None

    def start_select(self, attrs):
        if self._current_form is None:
            self.error(ParseError("start of SELECT before start of FORM"))
        if self._select is not None:
            self.error(ParseError("nested SELECTs"))
        if self._textarea is not None:
            self.error(ParseError("SELECT inside TEXTAREA"))
        d = {}
        for key, val in attrs:
            d[key] = val

        # SELECT doesn't *have* a "value" HTML attribute!
##         value = d.get("value")
##         if value is not None:
##             d["select_value"] = value

        self._select = d

    def end_select(self):
        if self._current_form is None:
            self.error(ParseError("end of SELECT before start of FORM"))
        if self._select is None:
            self.error(ParseError("end of SELECT before start"))
        self._select = None

    def do_option(self, attrs):
        if self._current_form is None:
            self.error(ParseError("OPTION outside of FORM"))
        if self._select is None:
            self.error(ParseError("OPTION outside of SELECT"))
        d = {}
        for key, val in attrs:
            d[key] = val

        self._option = self._select.copy()
        self._option.update(d)

    def start_textarea(self, attrs):
        if self._current_form is None:
            self.error(ParseError("start of TEXTAREA before start of FORM"))
        if self._textarea is not None:
            self.error(ParseError("nested TEXTAREAs"))
        if self._select is not None:
            self.error(ParseError("TEXTAREA inside SELECT"))
        d = {}
        for key, val in attrs:
            d[key] = val

        self._textarea = d

    def end_textarea(self):
        if self._current_form is None:
            self.error(ParseError("end of TEXTAREA before start of FORM"))
        if self._textarea is None:
            self.error(ParseError("end of TEXTAREA before start"))
        controls = self._current_form[2]
        controls.append(("textarea", self._textarea))
        self._textarea = None

    def handle_data(self, data):
        if self._option is not None:
            contents = string.strip(data)
            controls = self._current_form[2]
            if not self._option.has_key("value"):
                self._option["value"] = contents
            if not self._option.has_key("label"):
                self._option["label"] = contents
            # Hack: stuff text between OPTION tags into dict of HTML attrs
            self._option["contents"] = contents
            controls.append(("select", self._option))
            self._option = None
        elif self._textarea is not None:
            contents = string.strip(data)
            self._textarea["value"] = contents

    def do_button(self, attrs):
        if self._current_form is None:
            self.error(ParseError("start of BUTTON before start of FORM"))
        d = {}
        d["type"] = "submit"  # default
        for key, val in attrs:
            d[key] = val
        controls = self._current_form[2]

        type = d["type"]
        del d["type"]
        # we don't want to lose information, so use a type string that
        # doesn't clash with INPUT TYPE=SUBMIT or INPUT TYPE=RESET
        if type == "reset": type = "resetbutton"
        if type == "submit": type = "submitbutton"
        controls.append((type, d))

    def do_input(self, attrs):
        if self._current_form is None:
            self.error(ParseError("start of INPUT before start of FORM"))
        d = {}
        d["type"] = "text"  # default
        for key, val in attrs:
            d[key] = val
        controls = self._current_form[2]

        type = d["type"]
        del d["type"]
        controls.append((type, d))

    def do_isindex(self, attrs):
        if self._current_form is None:
            self.error(ParseError("start of ISINDEX before start of FORM"))
        d = {}
        for key, val in attrs:
            d[key] = val
        controls = self._current_form[2]

        type = "isindex"  # isindex doesn't have type HTML attribute
        if d.has_key("type"):
            del d["type"]  # shouldn't happen, but you never know
        controls.append((type, d))


class HTMLForm:
    """Represents a single HTML <form> ... </form> instance.

    A form consists of a sequence of controls that usually have names, and
    which can take on various values.

    An HTMLForm is only properly initialised after the fixup method has been
    called.  See ListControl.__doc__ for the reason this is required.  Usually,
    however, the constructor is not called directly; rather, the ParseFile or
    ParseResponse factory functions are used.  In that case, fixup is called
    for you:

    import ClientForm
    forms = ClientForm.ParseFile(html, base_uri)
    form = forms[0]
    assert form["query"].type == "text"
    form["query"] = "Python"
    response = urllib2.urlopen(form.click())

    Security note: Remember that any passwords you store in HTMLForm instances
    will be saved to disk in the clear if you pickle them (directly or
    indirectly).  The simplest solution to this is to avoid pickling HTMLForm
    objects.

    Public attributes:

    action: full (absolute URI) form action
    method: "GET" or "POST"
    enctype: form transfer encoding MIME type
    name: name of form (None if no name was specified)
    attrs: dictionary mapping other HTML form attributes to their values

    controls: list of Control instances; do not append to this list
     (instead, call form.new_control to make a Control and add it to the
     form, or control.add_to_form if you already have a Control instance)

    """
    def __init__(self, action, method="GET",
                 enctype="application/x-www-form-urlencoded",
                 name=None, attrs=None):
        """
        In the usual case, use ParseResponse (or ParseFile) to create new
        HTMLForm objects.

        action: full (absolute URI) form action
        method: "GET" or "POST"
        enctype: form transfer encoding MIME type
        name: name of form
        attrs: dictionary mapping all other HTML form attributes to their
         values

        """
        self.action = action
        self.method = method
        self.enctype = enctype
        self.name = name
        if attrs is not None:
            self.attrs = attrs.copy()
        else:
            self.attrs = {}
        self.controls = []

    def new_control(self, type, attrs,
                    ignore_unknown=False, select_default=False):
        """Adds a new control to the form.

        Note that controls representing lists of items are built up from
        controls holding only a single list item.  See ListControl.__doc__ for
        further information.

        type: type of control (see Control.__doc__ for a list)
        attrs: HTML attributes of control
        ignore_unknown: if true, use a dummy Control instance for controls of
         unknown type; otherwise, raise ValueError
        select_default: for RADIO and multiple-selection SELECT controls, pick
         the first item as the default if no 'selected' HTML attribute is
         present (this defaulting happens when the HTMLForm.fixup method is
         called)

        """
        type = string.lower(type)
        klass = TYPE2CLASS.get(type)
        if klass is None:
            if ignore_unknown:
                klass = IgnoreControl
            else:
                raise ValueError, "Unknown control type '%s'" % type

        a = attrs.copy()
        if issubclass(klass, ListControl):
            control = klass(type, a, select_default)
        else:
            control = klass(type, a)
        control.add_to_form(self)

    def fixup(self):
        """Normalise form after all controls have been added.

        This method should only be called once, after all controls have been
        added to the form.

        """
        for control in self.controls:
            control.fixup()

    def find_control(self, name=None, type=None, nr=0):
        """Locate some specific control within the form.

        At least one of the arguments must be supplied.  If no matching control
        is found, None is returned.

        If name is specified, then the control must have the indicated name.
        If type is specified then the control must have the specified type.  In
        addition to the types possible for <input> HTML tags, we also have
        "button", "textarea" and "select".  nr is the sequence number of the
        control with the indicated name and/or type (where 0 is the first).
        Note that control 0 is the first control matching the name and type
        arguments (if supplied); it is not the first control in the form.

        """
        if (name is None) and (type is None) and (nr is None):
            raise ValueError, \
                  "At least one of name, type and nr args must be supplied"
        if nr < 0:
            raise ValueError, "nr must be a positive integer"
        for control in self.controls:
            if name is not None:
                controlname = control.name
                if controlname is None:
                    continue
                if name != controlname:
                    continue
            if type is not None and type != control.type:
                continue
            if nr:
                nr = nr - 1
                continue
            return control
        return None

    def possible_values(self, name=None, type=None, nr=0):
        """Return a list of values for the specified control.

        See find_control.__doc__ for the meanings of the arguments.

        For controls that do not have a prescribed set of values, this returns
        an empty list.

        """
        ipt = self.find_control(name, type, nr)
        return ipt.possible_values()

    def toggle(self, name, value):
        """Toggle a named ListControl's value."""
        control = self.find_control(name)
        if control is None:
            raise ValueError, "control %s not found in form" % name
        if not hasattr(control, "toggle"):
            raise TypeError, \
                  "control %s does not have toggle-able values" % name
        control.toggle(value)

    def __setitem__(self, name, value):
        """Set the value(s) of named control."""
        control = self.find_control(name)
        if control is None:
            raise IndexError(name)
        try:
            control.value = value
        except AttributeError, e:
            raise ValueError, str(e)

    def __getitem__(self, name):
        """Get the value(s) of named control."""
        control = self.find_control(name)
        if control is None:
            raise IndexError(name)
        return control.value

    def items(self):
        """Return sequence of (key, value) pairs suitable for urlencoding.

        You probably want to use the click method instead, possibly with the
        items argument.

        Note that Python 2.0's urllib.urlencode is slightly broken: it only
        accepts a mapping, not a sequence of pairs, as an argument.  This
        messes up any ordering in the argument.  Use ClientForm.urlencode
        instead.

        """
        items = []
        for control in self.controls:
            items.extend(control.items())
        return items

    def make_request(self):
        """Return a urllib2.request object reflecting current form settings.

        You probably want to use the click method instead.

        Currently, ValueError is raised if the HTMLForm has method "POST" and
        an enctype that is not "application/x-www-form-urlencoded".

        """
        method = string.upper(self.method)
        if method == "GET":
            uri = "%s?%s" % (self.action, urlencode(self.items()))
            return urllib2.Request(uri)
        elif method == "POST":
            if self.enctype != "application/x-www-form-urlencoded":
                raise ValueError, \
                      "unknown form encoding type %s" % self.enctype
            req = urllib2.Request(self.action, urlencode(self.items()))
            req.add_header("Content-type", self.enctype)
            return req
        else:
            raise ValueError, "Unknown method '%s'" % method

    def click(self, name=None, coord=(1,1), nr=0, items=False):
        """Return request that would result from clicking on a control.

        Will click on the first clickable control, with the indicated name, if
        specified.  If nr is specified, the second (nr=1), third (nr=2) etc.
        submit control will be clicked.  Only some control types (INPUT/SUBMIT
        & BUTTON/SUBMIT buttons and IMAGEs) can be click()ed.  If no name or
        number is specified and there are no clickable controls, a request will
        be returned for the form in its current state.  You can optionally
        specify a coordinate to click at, which only makes a difference if you
        clicked on an image.

        ValueError is raised if a name is specified but no matching control is
        found.  ValueError is raised if the HTMLForm has method "POST" and an
        enctype attribute that is not "application/x-www-form-urlencoded".

        The request object is a urllib2.Request instance.

        If items is true, returns a list of (key, value) pairs instead of a
        request object.  You can pass this list to ClientForm.urlencode.  This
        is usually only useful if you're using urllib rather than urllib2 or
        ClientCookie.  It may also be useful if you want to manually tweak the
        items, but this should not be necessary.

        """
        # try to find first submit button to activate
        for control in self.controls:
            if not hasattr(control, "click"): continue
            if name is not None and control.name != name: continue
            if nr:
                nr = nr - 1
                continue
            return control.click(self, coord, items)
        if (name is not None) or (nr != 0):
            raise ValueError("No clickable control with name '%s' and nr %d" %
                             (name, nr))

        # no clickable controls, but no control was explicitly requested by
        # name, so return current state
        if items:
            return self.items()
        else:
            return self.make_request()

    def __str__(self):
        header = "%s %s %s" % (self.method, self.action, self.enctype)
        rep = [header]
        for control in self.controls:
            rep.append("  %s" % str(control))
        return "<%s>" % string.join(rep, "\n")


class Control:
    """An HTMLForm contains a sequence of Controls.

    The Controls in an HTMLForm can be got at via the HTMLForm.find_control
    method or the HTMLForm.controls attribute.

    A Control is only properly initialised after the fixup method has been
    called.  In fact, this is only strictly necessary for ListControl
    instances.  This is necessary because ListControls are built up from
    length-1 ListControls, and their initial value(s) can only be known after
    the sequence is complete.

    It is always allowable to assign a string to the value attribute (unless
    the control is disabled or should not be assigned to for some other
    reason).  Derived classes (in particular, ListControl instances whose
    multiple attribute is true) may also allow assignment of sequences.

    If the disabled attribute is true, this represents the state typically
    represented by browsers by `greying out' a control.  If the disabled
    attribute is true, the Control will raise AttributeError if an attempt is
    made to change its value.  In addition, the control will not be considered
    `successful' as defined by the W3C HTML 4 standard -- ie. its (name, value)
    pair will not be included in the output of the items method.  To enable a
    control, set the disabled attribute to a false value.  All controls have
    the disabled attribute, not only those that may have the disabled HTML
    attribute.

    If the name or value attributes are None, or if the control is disabled,
    the control is not successful.

    Public attributes:

    type: see the keys of the TYPE2CLASS dictionary in this module for the
     allowable values (readonly)
    name: name of control (readonly)
    value: current value of control (subclassess may allow a single value, a
     sequence of values, or either)
    disabled: disabled state

    """
    def __init__(self, type, attrs):
        """
        type: string describing type of control (see the keys of the TYPE2CLASS
         dictionary in this module for the allowable values)
        attrs: HTML attributes of control's HTML element

        """
        raise NotImplementedError

    def __setattr__(self, name, value):
        if name == "value":
            if self.disabled:
                raise AttributeError, "control '%s' is disabled" % self.name
            self.__dict__["_value"] = value
        elif name in ("name", "type"):
            raise AttributeError, "%s attribute is readonly" % name
        else:
            self.__dict__[name] = value

    def __getattr__(self, name):
        if name == "value":
            return self.__dict__["_value"]
        else:
            raise AttributeError, "%s instance has no attribute '%s'" % (
                self.__class__.__name__, name)

    def add_to_form(self, form):
        form.controls.append(self)

    def fixup(self):
        pass

    def possible_values(self):
        """Return a list of all values that the control can take.

        For controls that do not have a prescribed set of values, this returns
        an empty list.

        """
        return []

    def items(self):
        """Return list of (key, value) pairs suitable for passing to urlencode.

        Note that Python 2.0's urllib.urlencode is slightly broken: it only
        accepts a mapping, not a sequence of pairs, as an argument.  This
        messes up any ordering in the argument.  Use ClientForm.urlencode
        instead.

        """
        name = self.name
        value = self.value
        if name is None or value is None or self.disabled:
            return []
        return [(name, value)]

    def __str__(self):
        name = self.name
        value = self.value
        if name is None: name = "<None>"
        if value is None: value = "<None>"
        if self.disabled: dis = " (disabled)"
        else: dis = ""
        return "<%s(%s=%s)%s>" % (self.__class__.__name__, name, value, dis)


#---------------------------------------------------
class ScalarControl(Control):
    """Non-sequence control, having a single value.

    Additional public attributes:

    attrs: dictionary mapping the names of HTML attributes of the control
     to their values; only HTML attributes not accessible via the
     ScalarControl interface are found in the dictionary (so, for example,
     "name", "type" and "value" are never keys of the dictionary, but
     "alt" may be)

    """
    def __init__(self, type, attrs):
        self.__dict__["type"] = string.lower(type)
        self.__dict__["name"] = attrs.get("name")
        self._value = attrs.get("value")
        self.disabled = attrs.has_key("disabled")

        # Find the set of HTML attributes not already accessible from the
        # Control interface.
        a = self.attrs = attrs.copy()
        if a.has_key("type"): del a["type"]
        if self.name is not None: del a["name"]
        if self._value is not None: del a["value"]
        if self.disabled: del a["disabled"]

        self._clicked = False


#---------------------------------------------------
class TextControl(ScalarControl):
    """Textual input control.

    Covers:

    INPUT/TEXT
    INPUT/PASSWORD
    INPUT/FILE
    INPUT/HIDDEN
    TEXTAREA

    If the readonly attribute is true, the Control will raise AttributeError if
    an attempt is made to change its value.  To make a control writable, set
    the readonly attribute to a false value.

    Additional public attributes:

    readonly: readonly state

    """
    def __init__(self, type, attrs):
        ScalarControl.__init__(self, type, attrs)
        self.readonly = attrs.has_key("readonly")
        if self.readonly: del self.attrs["readonly"]
        #if self.type == "hidden": self.readonly = True
        if self._value is None:
            self._value = ""

    def __setattr__(self, name, value):
        if name == "value":
            if self.readonly:
                raise AttributeError, "control '%s' is readonly" % self.name
            if self.disabled:
                raise AttributeError, "control '%s' is disabled" % self.name
            self.__dict__["_value"] = value
        elif name in ("name", "type"):
            raise AttributeError, "%s attribute is readonly" % name
        else:
            self.__dict__[name] = value

    def __str__(self):
        name = self.name
        value = self.value
        if name is None: name = "<None>"
        if value is None: value = "<None>"

        infos = []
        if self.disabled: infos.append("disabled")
        if self.readonly: infos.append("readonly")
        info = string.join(infos, ", ")
        if info: info = " (%s)" % info

        return "<%s(%s=%s)%s>" % (self.__class__.__name__, name, value, info)


#---------------------------------------------------
class IsindexControl(ScalarControl):
    """ISINDEX control.

    ISINDEX is the odd-one-out of HTML form controls.  In fact, it isn't really
    part of regular HTML forms at all, since it predates them.  You're only
    allowed one ISINDEX per HTML document.  ISINDEX and regular form submission
    are mutually exclusive -- either submit the form, or the ISINDEX.

    Having said this, since ISINDEX controls may appear in forms (which is
    probably bad HTML), ParseFile / ParseResponse will include them in the
    HTMLForm instances it returns.  You can set the ISINDEX's value, as with
    any other control (but note that ISINDEX controls have no name, so you'll
    need to use the type argument of find_control!).  When you submit the form,
    the ISINDEX will not be successful (ie., no data will get returned to the
    server as a result of its presence).  If you want to submit the ISINDEX,
    rather than the form:

    isindex = form.find_control(type="isindex")
    url = urljoin(form.action, isindex.isindex_url())
    result = urllib2.urlopen(url)

    ISINDEX elements outside of FORMs are ignored.  If you want to submit one
    by hand, do it like so:

    url = urljoin(base_uri, "?"+urllib.quote_plus("text to submit"))
    result = urllib2.urlopen(url)

    XXX Umm, that's actually simpler... :-) this makes more sense in 0.1.x.

    """
    def __init__(self, type, attrs):
        ScalarControl.__init__(self, type, attrs)
        if self._value is None:
            self._value = ""

    def items(self):
        return []

    def isindex_url(self):
        """Return relative URL for ISINDEX submission"""
        # Instead of "foo=bar+baz", want "bar+baz".
        # This doesn't seem to be specified in HTML 4.01 spec. (ISINDEX is
        # deprecated in 4.01, but it should still say how to submit it).
        # Submission of ISINDEX is explained in the HTML 3.2 spec, though.
        return "?"+urllib.quote_plus(self.value)

    def __str__(self):
        value = self.value
        if value is None: value = "<None>"

        infos = []
        if self.disabled: infos.append("disabled")
        info = string.join(infos, ", ")
        if info: info = " (%s)" % info

        return "<%s(%s)%s>" % (self.__class__.__name__, value, info)


#---------------------------------------------------
class IgnoreControl(ScalarControl):
    """Control that we're not interested in.

    Covers:

    INPUT/RESET
    BUTTON/RESET
    BUTTON/BUTTON

    These controls are always unsuccessful, in the terminology of HTML 4 --
    ie., they never require any information to be returned to the server.

    BUTTON/BUTTON is used to generate events for script embedded in HTML.

    """
    def __setattr__(self, name, value):
        if name == "value":
            raise AttributeError, \
                  "control '%s' is ignored, hence read-only" % self.name
        elif name in ("name", "type"):
            raise AttributeError, "%s attribute is readonly" % name
        else:
            self.__dict__[name] = value


#---------------------------------------------------
class ListControl(Control):
    """Control representing a sequence of items.

    Covers:

    SELECT/OPTION
    INPUT/RADIO
    INPUT/CHECKBOX

    May always assign a string to value attribute (unless control is disabled).
    May assign a sequence to value attribute if controls accepts multiple
    values (SELECT with the "multiple" HTML attribute, and CHECKBOX).

    ListControl implements both list controls that take a single value and
    those that take multiple values.  RADIO lists always take a single
    value, CHECKBOXES always take multiple values, and SELECT takes
    multiple or single values, depending on the presence or absence of the
    "multiple" HTML attribute.

    Additional read-only public attribute: multiple.

    ListControls are built up by the parser from their component items by
    creating one ListControl per item, consolidating them into a single master
    ListControls held by the HTMLForm:

    -User calls form.new_control(...)
    -Form creates Control, and calls control.add_to_form(self).
    -Control looks for a Control with the same name and type in the form, and
    if it finds one, merges itself with that control by calling
    control.merge_control(self).  The first Control added to the form, of a
    particular name and type, is the only one that survives in the form.
    -Form calls control.fixup for all its controls.  ListControls in the form
    know they can now safely pick their default values.

    To create a ListControl without an HTMLForm, use control.merge_control.

    """
    def __init__(self, type, attrs, select_default=False,
                 called_as_base_class=False):
        """
        select_default: for RADIO and multiple-selection SELECT controls, pick
         the first item as the default if no 'selected' HTML attribute is
         present

        """
        if not called_as_base_class:
            raise NotImplementedError

        self.__dict__["type"] = string.lower(type)
        self.__dict__["name"] = attrs.get("name")
        self._value = attrs.get("value")
        self.disabled = attrs.has_key("disabled")

        # Find the set of HTML attributes not already accessible from the
        # Control interface.
        self._attrs = attrs.copy()
        if self._attrs.has_key("type"): del self._attrs["type"]
        if self.name is not None: del self._attrs["name"]
        if self._value is not None: del self._attrs["value"]
        if self.disabled: del self._attrs["disabled"]

        self._select_default = select_default
        self._clicked = False
        # Some list controls can have their default set only after all items
        # are known.  If so, self._value_is_set is false, and the self.fixup
        # method, called after all items have been added, sets the default.
        self._value_is_set = False

        self._attrs_list = [self._attrs]  # extended by .merge_control()

    def toggle(self, name):
        """Toggle named value."""
        if self.disabled:
            raise ValueError, "control '%s' is disabled" % self.name
        if not self.multiple:
            raise TypeError, \
                  "control %s does not have toggle-able values" % name

        # name and item_name here are values of the HTML attributes named
        # "value".  Example:
        #  <INPUT type=CHECKBOX name="cheeses" value="leicester"></INPUT>
        #  <INPUT type=CHECKBOX name="cheeses" value="cheddar"></INPUT>
        for i in range(len(self._menu)):
            item_name = self._menu[i]
            if name != item_name:
                continue
            self._selected[i] = not self._selected[i]
            return
        raise ValueError, "item not found in control"

    def find_item(self, name=None, nr=0):
        """Return the dictionary of attributes for a single item in the control.

        The items we're talking about here represent the possible values of a
        list control.  The HTML elements that describe these items are: OPTION
        for SELECT controls, INPUT for the rest.  These elements have HTML
        attributes that you may occasionally want to know about -- for example,
        the "alt" HTML attribute gives a text string describing the item
        (graphical browsers usually display this as a tooltip).

        The form of the dictionary of HTML attributes is identical to that of
        the attrs attribute of ScalarControl, as described in
        ScalarControl.__doc__.

        At least one of the arguments must be supplied.  If no matching item is
        found, None is returned.

        If name is specified, then the item must have the indicated name (note
        that this name is usually specified in HTML as the value of an HTML
        attribute named "value").  nr is the sequence number of the item with
        the indicated name (where 0 is the first).  Note that item 0 is the
        first item matching the name argument (if supplied); it is not
        necessarily the first item in the control.

        """
        if (name is None) and (nr is None):
            raise ValueError, \
                  "At least one of name and nr args must be supplied"
        for i in range(len(self._menu)):
            item_name = self._menu[i]
            if name != item_name:
                continue
            if nr:
                nr = nr - 1
                continue
            return self._attrs_list[i]
        return None

    def add_to_form(self, form):
        control = form.find_control(self.name, self.type)
        if control is None:
            Control.add_to_form(self, form)
        else:
            control.merge_control(self)

    def merge_control(self, control):
        assert isinstance(control, self.__class__)
        # merge menus
        self._menu.extend(control._menu)
        self._attrs_list.extend(control._attrs_list)
        if control.multiple:
            self._selected.extend(control._selected)
        else:
            if control._value_is_set:
                self._selected = control._selected
        if control._value_is_set:
            self._value_is_set = True

    def fixup(self):
        """
        ListControls are built up from component list items (which are also
        ListControls) during parsing.  This method should be called after all
        items have been added.  See ListControl.__doc__ for the reason this is
        required.

        """
        # Need to set default selection where no item was indicated as being
        # selected by the HTML:

        # CHECKBOX:
        #  Nothing should be selected.
        # SELECT/single, SELECT/multiple and RADIO:
        #  RFC 1866 (HTML 2.0): says first item should be selected.
        #  W3C HTML 4.01 Specification: says that client behaviour is
        #   undefined in this case.  For RADIO, exactly one must be selected,
        #   though which one is undefined.
        #  Both Netscape and Microsoft Internet Explorer (IE) choose first
        #   item for SELECT/single.  However, both IE5 and Mozilla (both 1.0
        #   and Firebird 0.6) leave all items unselected for RADIO and
        #   SELECT/multiple.

        # Since both Netscape and IE all choose the first item for
        # SELECT/single, we do the same.  OTOH, both Netscape and IE
        # leave SELECT/multiple with nothing selected, in violation of RFC 1866
        # (but not in violation of the W3C HTML 4 standard); the same is true
        # of RADIO (which *is* in violation of the HTML 4 standard).  We follow
        # RFC 1866 if the select_default attribute is set, and Netscape and IE
        # otherwise.  RFC 1866 and HTML 4 are always violated insofar as you
        # can deselect all items in a RadioControl.

        raise NotImplementedError

    def __getattr__(self, name):
        if name == "value":
            menu = self._menu
            if self.multiple:
                values = []
                for i in range(len(menu)):
                    if self._selected[i]: values.append(menu[i])
                return values
            else:
                return self._selected
        else:
            raise AttributeError, "%s instance has no attribute '%s'" % (
                self.__class__.__name__, name)

    def _single_set_value(self, value):
        if value is None:
            self._selected = value
        elif not isstringlike(value):
            raise AttributeError, ("Single selection list, must set a string "
                                   "or None")
        else:
            try:
                self._menu.index(value)
            except ValueError:
                raise AttributeError, "Illegal value %s" % repr(value)
            self._selected = value

    def _multiple_set_value(self, value):
        if value is None:
            value = []
        if isstringlike(value):
            value = [value]
        selected = [0]*len(self._selected)
        menu = self._menu

        for v in value:
            found = False
            for i in range(len(menu)):
                item_name = menu[i]
                if v == item_name:
                    selected[i] = True
                    found = True
                    break
            if not found:
                raise AttributeError, "Illegal value %s" % repr(v)
        self._selected = selected

    def __setattr__(self, name, value):
        if name == "value":
            if self.disabled:
                raise AttributeError, "control '%s' is disabled" % self.name
            if self.multiple:
                self._multiple_set_value(value)
            else:
                self._single_set_value(value)
        elif name in ("name", "type"):
            raise AttributeError, "%s attribute is readonly" % name
        else:
            self.__dict__[name] = value

    def possible_values(self):
        return copy(self._menu)

    def items(self):
        if self.disabled:
            return []

        if not self.multiple:
            return Control.items(self)

        control_name = self.name  # the name HTML attribute
        items = []
        for i in range(len(self._menu)):
            item_name = self._menu[i]  # the value HTML attribute
            if self._selected[i]:
                items.append((control_name, item_name))
        return items

    def __str__(self):
        menu = self._menu
        display = []
        for i in range(len(menu)):
            item_name = menu[i]
            if self.multiple:
                if self._selected[i]:
                    item_name = "*"+item_name
            else:
                if self._selected == item_name:
                    item_name = "*"+item_name
            display.append(item_name)

        if self.disabled: dis = " (disabled)"
        else: dis = ""
        return "<%s(%s=[%s])%s>" % (self.__class__.__name__,
                                    self.name,
                                    string.join(display, ", "),
                                    dis)


class RadioControl(ListControl):
    def __init__(self, type, attrs, select_default=False):
        ListControl.__init__(self, type, attrs, select_default,
                             called_as_base_class=True)
        self.multiple = False
        value = attrs.get("value", "on")
        self._menu = [value]
        checked = attrs.has_key("checked")
        if checked:
            self._value_is_set = True
            del self._attrs["checked"]
            self._selected = value
        else:
            self._selected = None

    def fixup(self):
        if not self._value_is_set:
            # no item explicitly selected
            assert self._selected is None
            if self._select_default:
                self._selected = self._menu[0]
            self._value_is_set = True


class CheckboxControl(ListControl):
    def __init__(self, type, attrs, select_default=False):
        ListControl.__init__(self, type, attrs, select_default,
                             called_as_base_class=True)
        self.multiple = True
        value = attrs.get("value", "on")
        self._menu = [value]
        checked = attrs.has_key("checked")
        if checked:
            del self._attrs["checked"]
        self._selected = [checked]
        self._value_is_set = True

    def fixup(self):
        # If no items were explicitly checked in HTML, that's how we must
        # leave it, so we have nothing to do here.
        assert self._value_is_set


class SelectControl(ListControl):
    # HTML attributes here are treated slightly from other list controls:
    # -The SELECT HTML attributes are stuffed into the same dictionary as
    #  the OPTION HTML attributes (the latter override the former).
    # -The content of each OPTION element is stuffed into the "contents"
    #  key of the dictionary.
    # After all this, the dictionary is passed to the SelectControl
    # constructor as the attrs argument, as usual.
    def __init__(self, type, attrs, select_default=False):
        ListControl.__init__(self, type, attrs, select_default,
                             called_as_base_class=True)
        self._label_map = None
        value = attrs["value"]
        self._menu = [value]
        selected = attrs.has_key("selected")
        if selected:
            self._value_is_set = True
            del self._attrs["selected"]
        if attrs.has_key("multiple"):
            self.multiple = True
            del self._attrs["multiple"]
            self._selected = [selected]
        else:
            self.multiple = False
            if selected:
                self._selected = value
            else:
                self._selected = None

    def _build_select_label_map(self):
        """Return an ordered mapping of labels to values.

        For example, if the HTML repreentation of the control is:

        <SELECT name=year>
         <OPTION value=0 label="2002">current year</OPTION>
         <OPTION value=1>2001</OPTION>
         <OPTION>2000</OPTION>
        </SELECT>

        this function will return a mapping like:

        {"2002": "0", "2001": "1", "2000": "2000"}

        Note that the value of the last OPTION defaults to its contents, as
        specified by RFC 1866, as do the labels of the second and third
        OPTIONs.

        The purpose of this is that the OPTION labels are sometimes much
        more meaningful, hence useful, than are the OPTION values.

        """
        alist = []
        for val in self.possible_values():
            attrs = self.find_item(val)
            alist.append((attrs["label"], val))
        return AList(alist)

    def fixup(self):
        if not self._value_is_set:
            # No item explicitly selected.
            if self.multiple:
                if self._select_default:
                    self._selected[0] = True
            else:
                assert self._selected is None
                self._selected = self._menu[0]
            self._value_is_set = True
        self._label_map = self._build_select_label_map()

    def possible_labels(self):
        self._label_map.set_inverted(True)
        try:
            r = map(lambda v, self=self: self._label_map[v],
                    self.possible_values())
        finally:
            self._label_map.set_inverted(False)
        return r

    def set_value_by_label(self, value):
        if self.disabled:
            raise AttributeError, "control '%s' is disabled" % self.name
        if self.multiple:
            if isstringlike(value):
                value = [value]
            value = map(lambda v, self=self: self._label_map[v], value)
            self._multiple_set_value(value)
        else:
            self._single_set_value(self._label_map[value])

    def get_value_as_label(self):
        menu = self._menu
        self._label_map.set_inverted(True)
        try:
            if self.multiple:
                values = []
                for i in range(len(menu)):
                    if self._selected[i]:
                        values.append(self._label_map[menu[i]])
                return values
            else:
                return self._label_map[self._selected]
        finally:
            self._label_map.set_inverted(False)

    def toggle_by_label(self, label):
        try:
            self.toggle(self._label_map[label])
        except KeyError:
            raise ValueError("no item has label %s" % label)


#---------------------------------------------------
class SubmitControl(ScalarControl):
    """
    Covers:

    INPUT/SUBMIT
    BUTTON/SUBMIT

    """
    def click(self, form, coord=(1,1), items=False):
        """Return a request object simulating a click on the control.

        Only some control types (submit buttons and images) can be click()ed.

        """
        self._clicked = coord
        if items:
            r = form.items()
        else:
            r = form.make_request()
        self._clicked = False
        return r

    def items(self):
        if self.disabled or not self._clicked:
            return []
        return Control.items(self)


#---------------------------------------------------
class ImageControl(SubmitControl):
    """
    Covers:

    INPUT/IMAGE

    """
    def items(self):
        clicked = self._clicked
        if self.disabled or not clicked:
            return []
        name = self.name
        if name is None: return []
        return [("%s.x" % name, clicked[0]),
                ("%s.y" % name, clicked[1])]


# aliases, just to make str(control) and str(form) clearer
class PasswordControl(TextControl): pass
class HiddenControl(TextControl): pass
class TextareaControl(TextControl): pass
class SubmitButtonControl(SubmitControl): pass

TYPE2CLASS = {
    "text": TextControl,
    "password": PasswordControl,
    "hidden": HiddenControl,
    "textarea": TextareaControl,

    "isindex": IsindexControl,

    "file": IgnoreControl,  # not implemented in 0.0.x

    "button": IgnoreControl,
    "resetbutton": IgnoreControl,
    "reset": IgnoreControl,

    "submit": SubmitControl,
    "submitbutton": SubmitButtonControl,
    "image": ImageControl,

    "radio": RadioControl,
    "checkbox": CheckboxControl,
    "select": SelectControl,
    }
